/*
 * ffmpeg-write support
 *
 * Partial Copyright (c) 2006-2009 Peter Schlaile, Stephan Henningsen (Zta).
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */


#ifdef WITH_FFMPEG
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/types.h>
#include <sys/wait.h>  // system()/WIFSIGNALED
#include <sys/stat.h> // struct stat, S_ISFIFO
#include <unistd.h>  // unlink()
#include <fcntl.h>   // mkfifo()/O_CREAT
#include <errno.h>   // mkfifo()/EEXIST
#include <signal.h>  // broken pipe handler.
#include <math.h>

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#if defined(_WIN32) && defined(_DEBUG) && !defined(__MINGW32__) && !defined(__CYGWIN__)
/* This does not seem necessary or present on MSVC 8, but may be needed in earlier versions? */
#if _MSC_VER < 1400
#include <stdint.h>
#endif
#endif
#if defined(WIN32) && (!(defined snprintf))
#define snprintf _snprintf
#endif

#include <ffmpeg/avformat.h>
#include <ffmpeg/avcodec.h>
#include <ffmpeg/rational.h>
#include <ffmpeg/swscale.h>
#include <ffmpeg/opt.h>

#include "BKE_writepipe.h"

#include "MEM_guardedalloc.h"
#include "BLI_blenlib.h"

#include "BKE_bad_level_calls.h"
#include "BKE_global.h"
#include "BKE_idprop.h"

#include "IMB_imbuf_types.h"
#include "IMB_imbuf.h"

#include "BSE_seqaudio.h"

#include "DNA_scene_types.h"
#include "blendef.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif



static RenderData *rd = NULL;

static char *output_filename = NULL;
static char *pipe_command = NULL;
static char *pipe_filename = NULL;
static FILE *pipe_file = NULL;

AVOutputFormat *fmt;
AVFormatContext *oc;
AVStream *audio_st, *video_st;
double audio_pts, video_pts;

/* 5 seconds stream duration */
#define STREAM_DURATION   5.0
#define STREAM_FRAME_RATE 25 /* 25 images/s */
#define STREAM_NB_FRAMES  ((int)(STREAM_DURATION * STREAM_FRAME_RATE))
#define STREAM_PIX_FMT PIX_FMT_YUV420P // default pix_fmt


static int sws_flags = SWS_BICUBIC;



/**************************************************************/
/* audio output */

float t, tincr, tincr2;
int16_t *samples;
uint8_t *audio_outbuf;
int audio_outbuf_size;
int audio_input_frame_size;

/*
 * add an audio output stream
 */
static AVStream *add_audio_stream(AVFormatContext *oc, int codec_id)
{
    AVCodecContext *c;
    AVStream *st;

    st = av_new_stream(oc, 1);
    if (!st) {
        fprintf(stderr, "Could not alloc stream\n");
        exit(1);
    }

    c = st->codec;
    c->codec_id = codec_id;
    c->codec_type = CODEC_TYPE_AUDIO;

    /* put sample parameters */
    c->bit_rate = 64000;
    c->sample_rate = 44100;
    c->channels = 2;
    return st;
}

static void open_audio(AVFormatContext *oc, AVStream *st)
{
    AVCodecContext *c;
    AVCodec *codec;

    c = st->codec;

    /* find the audio encoder */
    codec = avcodec_find_encoder(c->codec_id);
    if (!codec) {
        fprintf(stderr, "codec not found\n");
        exit(1);
    }

    /* open it */
    if (avcodec_open(c, codec) < 0) {
        fprintf(stderr, "could not open codec\n");
        exit(1);
    }

    /* init signal generator */
    t = 0;
    tincr = 2 * M_PI * 110.0 / c->sample_rate;
    /* increment frequency by 110 Hz per second */
    tincr2 = 2 * M_PI * 110.0 / c->sample_rate / c->sample_rate;

    audio_outbuf_size = 10000;
    audio_outbuf = av_malloc(audio_outbuf_size);

    /* ugly hack for PCM codecs (will be removed ASAP with new PCM
       support to compute the input frame size in samples */
    if (c->frame_size <= 1) {
        audio_input_frame_size = audio_outbuf_size / c->channels;
        switch(st->codec->codec_id) {
        case CODEC_ID_PCM_S16LE:
        case CODEC_ID_PCM_S16BE:
        case CODEC_ID_PCM_U16LE:
        case CODEC_ID_PCM_U16BE:
            audio_input_frame_size >>= 1;
            break;
        default:
            break;
        }
    } else {
        audio_input_frame_size = c->frame_size;
    }
    samples = av_malloc(audio_input_frame_size * 2 * c->channels);
}

/* prepare a 16 bit dummy audio frame of 'frame_size' samples and
   'nb_channels' channels */
static void get_audio_frame(int16_t *samples, int frame_size, int nb_channels)
{
    int j, i, v;
    int16_t *q;

    q = samples;
    for(j=0;j<frame_size;j++) {
        v = (int)(sin(t) * 10000);
        for(i = 0; i < nb_channels; i++)
            *q++ = v;
        t += tincr;
        tincr += tincr2;
    }
}

static void write_audio_frame(AVFormatContext *oc, AVStream *st)
{
    AVCodecContext *c;
    AVPacket pkt;
    av_init_packet(&pkt);

    c = st->codec;

    get_audio_frame(samples, audio_input_frame_size, c->channels);

    pkt.size= avcodec_encode_audio(c, audio_outbuf, audio_outbuf_size, samples);

    if (c->coded_frame->pts != AV_NOPTS_VALUE)
        pkt.pts= av_rescale_q(c->coded_frame->pts, c->time_base, st->time_base);
    pkt.flags |= PKT_FLAG_KEY;
    pkt.stream_index= st->index;
    pkt.data= audio_outbuf;

    /* write the compressed frame in the media file */
    if (av_interleaved_write_frame(oc, &pkt) != 0) {
        fprintf(stderr, "Error while writing audio frame\n");
        exit(1);
    }
}

static void close_audio(AVFormatContext *oc, AVStream *st)
{
    avcodec_close(st->codec);

    av_free(samples);
    av_free(audio_outbuf);
}

/**************************************************************/
/* video output */

AVFrame *picture, *tmp_picture;
uint8_t *video_outbuf;
int frame_count, video_outbuf_size;

/* add a video output stream */
static AVStream *add_video_stream(AVFormatContext *oc, int codec_id)
{
    AVCodecContext *c;
    AVStream *st;

    st = av_new_stream(oc, 0);
    if (!st) {
        fprintf(stderr, "Could not alloc stream\n");
        exit(1);
    }

    c = st->codec;
    c->codec_id = codec_id;
    c->codec_type = CODEC_TYPE_VIDEO;

    /* put sample parameters */
    c->bit_rate = 400000;
    /* resolution must be a multiple of two */
    c->width = 352;
    c->height = 288;
    /* time base: this is the fundamental unit of time (in seconds) in terms
       of which frame timestamps are represented. for fixed-fps content,
       timebase should be 1/framerate and timestamp increments should be
       identically 1. */
    c->time_base.den = STREAM_FRAME_RATE;
    c->time_base.num = 1;
    c->gop_size = 12; /* emit one intra frame every twelve frames at most */
    c->pix_fmt = STREAM_PIX_FMT;
    if (c->codec_id == CODEC_ID_MPEG2VIDEO) {
        /* just for testing, we also add B frames */
        c->max_b_frames = 2;
    }
    if (c->codec_id == CODEC_ID_MPEG1VIDEO){
        /* Needed to avoid using macroblocks in which some coeffs overflow.
           This does not happen with normal video, it just happens here as
           the motion of the chroma plane does not match the luma plane. */
        c->mb_decision=2;
    }
    // some formats want stream headers to be separate
    if(oc->oformat->flags & AVFMT_GLOBALHEADER)
        c->flags |= CODEC_FLAG_GLOBAL_HEADER;

    return st;
}

static AVFrame *alloc_picture(int pix_fmt, int width, int height)
{
    AVFrame *picture;
    uint8_t *picture_buf;
    int size;

    picture = avcodec_alloc_frame();
    if (!picture)
        return NULL;
    size = avpicture_get_size(pix_fmt, width, height);
    picture_buf = av_malloc(size);
    if (!picture_buf) {
        av_free(picture);
        return NULL;
    }
    avpicture_fill((AVPicture *)picture, picture_buf,
                   pix_fmt, width, height);
    return picture;
}

static void open_video(AVFormatContext *oc, AVStream *st)
{
    AVCodec *codec;
    AVCodecContext *c;

    c = st->codec;

    /* find the video encoder */
    codec = avcodec_find_encoder(c->codec_id);
    if (!codec) {
        fprintf(stderr, "codec not found\n");
        exit(1);
    }

    /* open the codec */
    if (avcodec_open(c, codec) < 0) {
        fprintf(stderr, "could not open codec\n");
        exit(1);
    }

    video_outbuf = NULL;
    if (!(oc->oformat->flags & AVFMT_RAWPICTURE)) {
        /* allocate output buffer */
        /* XXX: API change will be done */
        /* buffers passed into lav* can be allocated any way you prefer,
           as long as they're aligned enough for the architecture, and
           they're freed appropriately (such as using av_free for buffers
           allocated with av_malloc) */
        video_outbuf_size = 200000;
        video_outbuf = av_malloc(video_outbuf_size);
    }

    /* allocate the encoded raw picture */
    picture = alloc_picture(c->pix_fmt, c->width, c->height);
    if (!picture) {
        fprintf(stderr, "Could not allocate picture\n");
        exit(1);
    }

    /* if the output format is not YUV420P, then a temporary YUV420P
       picture is needed too. It is then converted to the required
       output format */
    tmp_picture = NULL;
    if (c->pix_fmt != PIX_FMT_YUV420P) {
        tmp_picture = alloc_picture(PIX_FMT_YUV420P, c->width, c->height);
        if (!tmp_picture) {
            fprintf(stderr, "Could not allocate temporary picture\n");
            exit(1);
        }
    }
}

/* prepare a dummy image */
static void fill_yuv_image(AVFrame *pict, int frame_index, int width, int height)
{
    int x, y, i;

    i = frame_index;

    /* Y */
    for(y=0;y<height;y++) {
        for(x=0;x<width;x++) {
            pict->data[0][y * pict->linesize[0] + x] = x + y + i * 3;
        }
    }

    /* Cb and Cr */
    for(y=0;y<height/2;y++) {
        for(x=0;x<width/2;x++) {
            pict->data[1][y * pict->linesize[1] + x] = 128 + y + i * 2;
            pict->data[2][y * pict->linesize[2] + x] = 64 + x + i * 5;
        }
    }
}

static void write_video_frame(AVFormatContext *oc, AVStream *st)
{
    int out_size, ret;
    AVCodecContext *c;
    static struct SwsContext *img_convert_ctx;

    c = st->codec;

    if (frame_count >= STREAM_NB_FRAMES) {
        /* no more frame to compress. The codec has a latency of a few
           frames if using B frames, so we get the last frames by
           passing the same picture again */
    } else {
        if (c->pix_fmt != PIX_FMT_YUV420P) {
            /* as we only generate a YUV420P picture, we must convert it
               to the codec pixel format if needed */
            if (img_convert_ctx == NULL) {
                img_convert_ctx = sws_getContext(c->width, c->height,
                                                 PIX_FMT_YUV420P,
                                                 c->width, c->height,
                                                 c->pix_fmt,
                                                 sws_flags, NULL, NULL, NULL);
                if (img_convert_ctx == NULL) {
                    fprintf(stderr, "Cannot initialize the conversion context\n");
                    exit(1);
                }
            }
            fill_yuv_image(tmp_picture, frame_count, c->width, c->height);
            sws_scale(img_convert_ctx, tmp_picture->data, tmp_picture->linesize,
                      0, c->height, picture->data, picture->linesize);
        } else {
            fill_yuv_image(picture, frame_count, c->width, c->height);
        }
    }


    if (oc->oformat->flags & AVFMT_RAWPICTURE) {
        /* raw video case. The API will change slightly in the near
           futur for that */
        AVPacket pkt;
        av_init_packet(&pkt);

        pkt.flags |= PKT_FLAG_KEY;
        pkt.stream_index= st->index;
        pkt.data= (uint8_t *)picture;
        pkt.size= sizeof(AVPicture);

        ret = av_interleaved_write_frame(oc, &pkt);
    } else {
        /* encode the image */
        out_size = avcodec_encode_video(c, video_outbuf, video_outbuf_size, picture);
        /* if zero size, it means the image was buffered */
        if (out_size > 0) {
            AVPacket pkt;
            av_init_packet(&pkt);

            if (c->coded_frame->pts != AV_NOPTS_VALUE)
                pkt.pts= av_rescale_q(c->coded_frame->pts, c->time_base, st->time_base);
            if(c->coded_frame->key_frame)
                pkt.flags |= PKT_FLAG_KEY;
            pkt.stream_index= st->index;
            pkt.data= video_outbuf;
            pkt.size= out_size;

            /* write the compressed frame in the media file */
            ret = av_interleaved_write_frame(oc, &pkt);
        } else {
            ret = 0;
        }
    }
    if (ret != 0) {
        fprintf(stderr, "Error while writing video frame\n");
        exit(1);
    }
    frame_count++;
}

static void close_video(AVFormatContext *oc, AVStream *st)
{
    avcodec_close(st->codec);
    av_free(picture->data[0]);
    av_free(picture);
    if (tmp_picture) {
        av_free(tmp_picture->data[0]);
        av_free(tmp_picture);
    }
    av_free(video_outbuf);
}




static void make_output_filename ( char **prefix, char* ext, int width, int height )
{
	char extra[32];

	*prefix = malloc( FILE_MAXDIR+FILE_MAXFILE );
	strcpy(*prefix, G.scene->r.pic);
	//replace(*prefix, "//", BLI_get_project_path() );
	BLI_convertstringcode(*prefix, G.sce);
	//BLI_make_existing_file(*prefix);

	sprintf(extra, "%s%ix%i.%04d-%04d%s", (BLI_is_dir(*prefix)?"":"."), width, height, (G.scene->r.sfra), (G.scene->r.efra), ext );

       	strcat(*prefix, extra);
}


static void make_pipe_command ( char **command,
				const char* user_command,
				const char *output_filename )
{
	*command = malloc( FILE_MAXDIR+FILE_MAXFILE );

	strcpy( *command, user_command );
	//fprintf(stderr,"1 command=[%s]\n", *command);

	// NOTE:  For some weird reason the BLI_convertstringcode does
	// the following invalid conversion (note final /render.log):

	// BLI_convertstringcode(*command, G.sce);
	// "/opt/ffmpeg/bin/ffmpeg -f yuv4mpegpipe -i - -acodec -libfaac -ab 192k -vcodec libx264 -vpre hq -crf 20 -level 41 -f mp4 % 2 > //render.log
	// "/opt/ffmpeg/bin/ffmpeg -f yuv4mpegpipe -i - -acodec -libfaac -ab 192k -vcodec libx264 -vpre hq -crf 20 -level 41 -f mp4 % 2 > /render.log]

	// ...and since I'm only intersted in the "//" expansion, I do it manually:
	while ( BLI_strreplace( command, "///", G.scene->r.pic ) );        // replace "///" with render output path
	while ( BLI_strreplace( command, "//", BLI_get_project_path() ) ); // replace "//" with project path.

	//fprintf(stderr,"2 command=[%s]\n", *command);
	while ( BLI_strreplace( command, FFMPEG_PIPE_KEY_OUTPUT_FILENAME, output_filename ) );

	//fprintf(stderr,"3 command=[%s]\n", *command);
}




void end_pipe( void );

static void broken_pipe_handler ( int sig )
{
	error("Broken pipe");
	if ( !G.afbreek ) {
		G.afbreek = 1;  // Abort render (if not already aborted).
		end_pipe();
	}
}



void start_pipe ( struct RenderData *rd_, int rectx, int recty )
{
	AVFormatContext	*avformatctx;
	AVOutputFormat	*fmt;
	char		*ext;

	char		pipe_filename[10]; // ffmpeg-style pipe filename: "pipe:n", n being file descriptor number.

//	fprintf(stderr, "Starting pipe... pipe_command=[%s]\n", G.scene->r.ffcodecdata.pipe_command );
	fprintf(stderr, "Starting pipe...\n");

	rd = rd_;




	// Make correct output filename and command.
	make_output_filename( &output_filename, ext, rectx, recty );
	make_pipe_command( &pipe_command, G.scene->r.ffcodecdata.pipe_command, output_filename );


    // initialize libavcodec, and register all codecs and formats
    av_register_all();

//    fmt = guess_format(NULL, filename, NULL);
    fmt = guess_format("avi", NULL, NULL);
//    fmt = guess_format("yuv4mpegpipe", NULL, NULL);
//    if (!fmt) {
//        printf("Could not deduce output format from file extension: using MPEG.\n");
//        fmt = guess_format("mpeg", NULL, NULL);
//    }
    if (!fmt) {
        error("Could not find suitable output format");
	G.afbreek = 1;
	return;
    }

    // allocate the output media context.
    oc = avformat_alloc_context();
    if (!oc) {
        error(stderr, "Memory error");
	G.afbreek = 1;
	return;
    }
    oc->oformat = fmt;


    // add the audio and video streams using the default format codecs and initialize the codecs
    video_st = NULL;
    audio_st = NULL;
    if (fmt->video_codec != CODEC_ID_NONE) {
        video_st = add_video_stream(oc, fmt->video_codec);
    } else {
        error("Note: No video stream in format.\n");
    }
    if (fmt->audio_codec != CODEC_ID_NONE) {
        audio_st = add_audio_stream(oc, fmt->audio_codec);
    } else {
        error("Note: No audio stream in format.\n");
    }

    // set the output parameters (must be done even if no parameters).
    if (av_set_parameters(oc, NULL) < 0) {
	error("Invalid output format parameters");
	G.afbreek = 1;
	return;
    }

    // now that all the parameters are set, we can open the audio and
    // video codecs and allocate the necessary encode buffers
    if (video_st) open_video(oc, video_st);
    if (audio_st) open_audio(oc, audio_st);


/*
    // open the output file, if needed
    if (!(fmt->flags & AVFMT_NOFILE)) {
        if (url_fopen(&oc->pb, pipe_filename, URL_WRONLY) < 0) {
            fprintf(stderr, "Could not open '%s'\n", pipe_filename);
            exit(1);
        }
    }
*/


	signal(SIGPIPE, broken_pipe_handler);

	pipe_file = popen( pipe_command, "w" ); // Open writable pipe FILE* from blender to stdin of command.
	if ( pipe_file == NULL ) {
			G.afbreek = 1;
			error("Could not open pipe to command or failed executing command: %s", strerror(errno));
			return;
	}
	sprintf(pipe_filename, "pipe:%i", fileno(pipe_file));
	snprintf(oc->filename, sizeof(oc->filename), "%s", pipe_filename);

        dump_format(oc, 0, pipe_filename, 1);

	fprintf(stderr, "Starting output to pipe: %ix%i ...\n"
			"command=[%s]\n"
			"pipe_filename=[%s], output_filename=[%s]\n",
			rectx, recty,
			pipe_command,
			pipe_filename, output_filename );


    // write the stream header, if any
    av_write_header(oc);


}


void append_pipe ( int frame, int *pixels, int rectx, int recty )
{

/*
	char filename[512];
	FILE* file;

	// Dump raw pixel data in separate files pr. frame for debugging.
	sprintf(filename, "/home/stephan/video-sjov/render/frame-%ix%i-%04i.raw", rectx, recty, frame );
	file = fopen( filename, "w");
	fwrite( pixels, rectx*recty, sizeof(int), file );
	fclose(file);
*/
//	write_audio_frames();
//	write_video_frame( generate_video_frame((unsigned char*) pixels) );

    for(;;) {
        // compute current audio and video time
        if (audio_st)
            audio_pts = (double)audio_st->pts.val * audio_st->time_base.num / audio_st->time_base.den;
        else
            audio_pts = 0.0;

        if (video_st)
            video_pts = (double)video_st->pts.val * video_st->time_base.num / video_st->time_base.den;
        else
            video_pts = 0.0;

        if ((!audio_st || audio_pts >= STREAM_DURATION) &&
            (!video_st || video_pts >= STREAM_DURATION))
            break;

        // write interleaved audio and video frames
        if (!video_st || (video_st && audio_st && audio_pts < video_pts)) {
            write_audio_frame(oc, audio_st);
        } else {
            write_video_frame(oc, video_st);
        }
    }
}


void end_pipe ( void )
{
	int i;

	fprintf(stderr, "Closing pipe... %i\n", fileno(pipe_file));

    // write the trailer, if any.  the trailer must be written
    // before you close the CodecContexts open when you wrote the
    // header; otherwise write_trailer may try to use memory that
    // was freed on av_codec_close() */
    av_write_trailer(oc);

    // close each codec
    if (video_st) close_video(oc, video_st);
    if (audio_st) close_audio(oc, audio_st);

    // free the streams
    for(i = 0; i < oc->nb_streams; i++) {
        av_freep(&oc->streams[i]->codec);
        av_freep(&oc->streams[i]);
    }

    if (!(fmt->flags & AVFMT_NOFILE)) {
        // close the output file
        url_fclose(oc->pb);
    }

    // free the stream
    av_free(oc);

	if ( pipe_filename != NULL ) 	{ free( pipe_filename );	pipe_filename = NULL; }
	if ( pipe_file != NULL ) 	{ pclose(pipe_file);		pipe_file = NULL; }
	if ( output_filename != NULL )	{ free( output_filename );	output_filename = NULL; }
	if ( pipe_command != NULL )	{ free( pipe_command );		pipe_command = NULL; }
}




#endif




